# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Explainer Manager class."""

import json
import pickle
import warnings
from pathlib import Path
from typing import Any, List, Optional

import numpy as np
import pandas as pd
import shap
from raiutils.data_processing import convert_to_list
from responsibleai._interfaces import (FeatureImportance, ModelExplanationData,
                                       PrecomputedExplanations,
                                       TextFeatureImportance)
from responsibleai._internal.constants import ExplainerManagerKeys as Keys
from responsibleai._internal.constants import (ListProperties, ManagerNames,
                                               Metadata)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.exceptions import UserConfigValidationException
from responsibleai.managers.base_manager import BaseManager

from responsibleai_text.common.constants import (ModelTask,
                                                 QuestionAnsweringFields,
                                                 Tokens)
from responsibleai_text.utils.question_answering import QAPredictor

CONTEXT = QuestionAnsweringFields.CONTEXT
QUESTIONS = QuestionAnsweringFields.QUESTIONS
SEP = Tokens.SEP
SPARSE_NUM_FEATURES_THRESHOLD = 1000
IS_RUN = 'is_run'
IS_ADDED = 'is_added'
CLASSES = 'classes'
U_EVALUATION_EXAMPLES = '_evaluation_examples'
FEATURES = 'features'
META_JSON = Metadata.META_JSON
MODEL = Metadata.MODEL
EXPLANATION = '_explanation'
TASK_TYPE = '_task_type'


class ExplainerManager(BaseManager):

    """Defines the ExplainerManager for explaining a text-based model."""

    def __init__(self, model: Any, evaluation_examples: pd.DataFrame,
                 target_column: str,
                 task_type: str,
                 classes: Optional[List] = None):
        """Creates an ExplainerManager object.

        :param model: The model to explain.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param evaluation_examples: A matrix of feature vector
            examples (# examples x # features) on which to explain the
            model's output, with an additional label column.
        :type evaluation_examples: pandas.DataFrame
        :param target_column: The name of the label column or list of columns.
            This is a list of columns for multilabel models.
        :type target_column: str or list[str]
        :param task_type: The task to run.
        :type task_type: str
        :param classes: Class names as a list of strings.
            The order of the class names should match that of the model
            output.  Only required if explaining classifier.
        :type classes: list
        """
        self._model = model
        self._target_column = target_column
        if not isinstance(target_column, list):
            target_column = [target_column]
        self._evaluation_examples = \
            evaluation_examples.drop(columns=target_column)
        self._is_run = False
        self._is_added = False
        self._features = list(self._evaluation_examples.columns)
        self._classes = classes
        self._explanation = None
        self._task_type = task_type

    def add(self):
        """Add an explainer to be computed later."""
        if self._model is None:
            raise UserConfigValidationException(
                'Model is required for model explanations')

        if self._is_added:
            warnings.warn(("DUPLICATE-EXPLAINER-CONFIG: Ignoring. "
                           "Explanation has already been added, "
                           "currently limited to one explainer type."),
                          UserWarning)
            return
        self._is_added = True

    def compute(self):
        """Creates an explanation by running the explainer on the model."""
        if not self._is_added:
            return
        if self._is_run:
            return
        if self._is_classification_task:
            if hasattr(self._model, 'predict_proba'):
                # use model-agnostic simple tokenizer
                masker = shap.maskers.Text()
                explainer = shap.Explainer(self._model.predict_proba,
                                           masker)
            else:
                explainer = shap.Explainer(self._model)
            eval_examples = self._evaluation_examples.iloc[:, 0].tolist()
            self._explanation = explainer(eval_examples)
        elif self._task_type == ModelTask.QUESTION_ANSWERING:
            qa_predictor = QAPredictor(self._model)
            qa_start = qa_predictor.predict_qa_start
            qa_start.__func__.output_names = qa_predictor.output_names
            explainer = shap.Explainer(qa_start, self._model.tokenizer)
            context = self._evaluation_examples[CONTEXT]
            questions = self._evaluation_examples[QUESTIONS]
            eval_examples = []
            for context, question in zip(context, questions):
                eval_examples.append(question + SEP + context)
            self._explanation = explainer(eval_examples)
        else:
            raise ValueError("Unknown task type: {}".format(self._task_type))

        self._is_run = True

    def get(self):
        """Get the computed explanation.

        Must be called after add and compute methods.

        :return: The computed explanations.
        :rtype:
            list[interpret_community.explanation.explanation.BaseExplanation]
        """
        if self._explanation:
            return [self._explanation]
        else:
            return []

    def list(self):
        """List information about the ExplainerManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        if self._explanation:
            props[Keys.IS_COMPUTED] = True
        else:
            props[Keys.IS_COMPUTED] = False
        return props

    def get_data(self):
        """Get explanation data

        :return: A array of ModelExplanationData.
        :rtype: List[ModelExplanationData]
        """
        return [self._get_interpret(i) for i in self.get()]

    @property
    def _is_multilabel_task(self):
        """Check if the task is a multilabel classification task.

        :return: True if the task is a multilabel classification task.
        :rtype: bool
        """
        return self._task_type == ModelTask.MULTILABEL_TEXT_CLASSIFICATION

    @property
    def _is_classification_task(self):
        """Check if the task is a classification task.

        :return: True if the task is a classification task.
        :rtype: bool
        """
        is_onelabel_task = self._task_type == ModelTask.TEXT_CLASSIFICATION
        is_multilabel_task = self._is_multilabel_task
        return is_onelabel_task or is_multilabel_task

    def _get_interpret(self, explanation):
        interpretation = ModelExplanationData()

        try:
            importances = FeatureImportance()
            features, scores, intercept = self._compute_global_importances(
                explanation)
            importances.featureNames = features
            importances.scores = scores
            importances.intercept = intercept
            text_feature_importances = self._compute_text_feature_importances(
                explanation)
            precomputedExplanations = PrecomputedExplanations()
            precomputedExplanations.globalFeatureImportance = importances
            precomputedExplanations.textFeatureImportance = \
                text_feature_importances
            interpretation.precomputedExplanations = precomputedExplanations
        except Exception as ex:
            raise ValueError(
                "Unsupported explanation type") from ex
        return interpretation

    def _compute_global_importances(self, explanation):
        """Compute global feature importances.

        :param explanation: The explanation.
        :type explanation: shap.Explanation
        :return: The feature names, scores, and intercept.
        :rtype: tuple[list[str], list[float], float]
        """
        is_classif_task = self._is_classification_task
        if is_classif_task:
            global_exp = explanation[:, :, :].mean(0)
            features = convert_to_list(global_exp.feature_names)
            scores = convert_to_list(np.abs(global_exp.values).mean(1))
            intercept = global_exp.base_values.mean(0)
        elif self._task_type == ModelTask.QUESTION_ANSWERING:
            flattened_features = explanation._flatten_feature_names()
            scores = []
            features = []
            for key in flattened_features.keys():
                features.append(key)
                token_importances = []
                for importances in flattened_features[key]:
                    token_importances.append(np.mean(np.abs(importances)))
                scores.append(np.mean(token_importances))
                base_values = [
                    base_values.mean()
                    for base_values in explanation.base_values]
                intercept = sum(base_values) / len(base_values)
        else:
            raise ValueError("Unknown task type: {}".format(self._task_type))

        return features, scores, intercept

    def _compute_text_feature_importances(self, explanation):
        """Compute the text feature importances.

        :param explanation: The explanation.
        :type explanation: shap.Explanation
        :return: The text importances and corresponding tokens.
        :rtype: tuple[list[str], list[float], float]
        """
        text_feature_importances = []
        is_classif_task = self._is_classification_task
        for instance in explanation:
            text_feature_importance = TextFeatureImportance()
            if is_classif_task:
                text_feature_importance.localExplanations = \
                    instance.values.tolist()
                text_feature_importance.text = instance.data
            elif self._task_type == ModelTask.QUESTION_ANSWERING:
                # TODO: This is a bit more complicated, as it's
                # a map of importances for each token from question
                # to answer and the other way around.
                continue
            else:
                raise ValueError("Unknown task type: {}".format(
                    self._task_type))
            text_feature_importances.append(text_feature_importance)
        return text_feature_importances

    @property
    def name(self):
        """Get the name of the explainer manager.

        :return: The name of the explainer manager.
        :rtype: str
        """
        return ManagerNames.EXPLAINER

    def _save(self, path):
        """Save the ExplainerManager to the given path.

        :param path: The directory path to save the ExplainerManager to.
        :type path: str
        """
        top_dir = Path(path)
        top_dir.mkdir(parents=True, exist_ok=True)
        if self._is_added:
            directory_manager = DirectoryManager(parent_directory_path=path)
            data_directory = directory_manager.create_data_directory()

            # save the explanation
            if self._explanation:
                with open(data_directory / ManagerNames.EXPLAINER, 'wb') as f:
                    pickle.dump(self._explanation, f)

            meta = {IS_RUN: self._is_run,
                    IS_ADDED: self._is_added}
            with open(data_directory / META_JSON, 'w') as file:
                json.dump(meta, file)

    @staticmethod
    def _load(path, rai_insights):
        """Load the ExplainerManager from the given path.

        :param path: The directory path to load the ExplainerManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The ExplainerManager manager after loading.
        :rtype: ExplainerManager
        """
        # create the ExplainerManager without any properties using the __new__
        # function, similar to pickle
        inst = ExplainerManager.__new__(ExplainerManager)

        all_cf_dirs = DirectoryManager.list_sub_directories(path)
        if len(all_cf_dirs) != 0:
            directory_manager = DirectoryManager(
                parent_directory_path=path,
                sub_directory_name=all_cf_dirs[0])
            data_directory = directory_manager.get_data_directory()

            with open(data_directory / META_JSON, 'r') as meta_file:
                meta = meta_file.read()
            meta = json.loads(meta)
            inst.__dict__['_' + IS_RUN] = meta[IS_RUN]
            inst.__dict__['_' + IS_ADDED] = meta[IS_ADDED]

            inst.__dict__[EXPLANATION] = None
            explanation_path = data_directory / ManagerNames.EXPLAINER
            if explanation_path.exists():
                with open(explanation_path, 'rb') as f:
                    explanation = pickle.load(f)
                inst.__dict__[EXPLANATION] = explanation
        else:
            inst.__dict__['_' + IS_RUN] = False
            inst.__dict__['_' + IS_ADDED] = False
            inst.__dict__[EXPLANATION] = None

        inst.__dict__['_' + MODEL] = rai_insights.model
        inst.__dict__['_' + CLASSES] = rai_insights._classes
        target_column = rai_insights.target_column
        if not isinstance(target_column, list):
            target_column = [target_column]
        test = rai_insights.test.drop(columns=target_column)
        inst.__dict__[U_EVALUATION_EXAMPLES] = test
        inst.__dict__['_' + FEATURES] = list(test.columns)
        inst.__dict__[TASK_TYPE] = rai_insights.task_type

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from common_text_utils import (ANSWERS, COVID19_EVENTS_LABELS, EMOTION,
                               create_multilabel_pipeline,
                               create_question_answering_pipeline,
                               create_text_classification_pipeline,
                               load_covid19_emergency_event_dataset,
                               load_emotion_dataset, load_squad_dataset)
from rai_text_insights_validator import validate_rai_text_insights

from responsibleai_text import ModelTask, RAITextInsights


class TestRAITextInsights(object):

    def test_rai_insights_emotion_classification(self):
        data = load_emotion_dataset()
        pred = create_text_classification_pipeline()
        task_type = ModelTask.TEXT_CLASSIFICATION
        run_rai_insights(pred, data, data[:3], EMOTION, task_type)

    def test_rai_insights_question_answering(self):
        data = load_squad_dataset()
        pred = create_question_answering_pipeline()
        task_type = ModelTask.QUESTION_ANSWERING
        run_rai_insights(pred, data, data[:5], ANSWERS, task_type)

    def test_rai_insights_multilabel(self):
        data = load_covid19_emergency_event_dataset()
        pred = create_multilabel_pipeline()
        task_type = ModelTask.MULTILABEL_TEXT_CLASSIFICATION
        labels = COVID19_EVENTS_LABELS
        run_rai_insights(pred, data, data[:5], labels, task_type)


def run_rai_insights(model, train_data, test_data,
                     target_column, task_type):
    if task_type == ModelTask.TEXT_CLASSIFICATION:
        classes = train_data[target_column].unique()
        classes.sort()
    else:
        classes = None
    rai_insights = RAITextInsights(model, test_data,
                                   target_column,
                                   task_type=task_type,
                                   classes=classes)
    rai_insights.explainer.add()
    if task_type != ModelTask.QUESTION_ANSWERING:
        rai_insights.error_analysis.add()
    rai_insights.compute()
    rai_insights.get_data()
    # Validate
    validate_rai_text_insights(
        rai_insights, classes, test_data,
        target_column, task_type)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Responsible AI Text SDK package."""

from responsibleai_text.common.constants import ModelTask
from responsibleai_text.rai_text_insights import RAITextInsights

from .version import name, version

__name__ = name
__version__ = version

__all__ = ['ModelTask', 'RAITextInsights']



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Implementation of RAI Text Insights API."""

from responsibleai_text.rai_text_insights.rai_text_insights import \
    RAITextInsights

__all__ = ['RAITextInsights']



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from enum import Enum


class ModelTask(str, Enum):
    """Provide model task constants.

    Can be 'text_classification', 'sentiment_analysis',
    'question_answering', 'entailment', 'summarizations'
    or 'unknown'.
    """

    TEXT_CLASSIFICATION = 'text_classification'
    MULTILABEL_TEXT_CLASSIFICATION = 'multilabel_text_classification'
    SENTIMENT_ANALYSIS = 'sentiment_analysis'
    QUESTION_ANSWERING = 'question_answering'
    ENTAILMENT = 'entailment'
    SUMMARIZATIONS = 'summarizations'
    UNKNOWN = 'unknown'


class Tokens(object):
    """Provide tokens related to text processing."""

    SEP = "[SEP]"


class QuestionAnsweringFields(object):
    """Provide fields related to question answering task."""

    QUESTIONS = "questions"
    QUESTION = "question"
    CONTEXT = "context"
    ANSWERS = "answers"



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the text feature extraction utilities."""

import re
from typing import List, Union

import pandas as pd
import spacy
from negspacy.termsets import termset
from nlp_feature_extractors import attribute_extractors as exts
from tqdm import tqdm

from responsibleai_text.common.constants import (ModelTask,
                                                 QuestionAnsweringFields)

nlp = None


def extract_features(text_dataset: pd.DataFrame,
                     target_column: Union[str, List], task_type: str):
    '''Extract tabular data features from the text dataset.

    :param text_dataset: A pandas dataframe containing the text data.
    :type text_dataset: pandas.DataFrame
    :param target_column: The name of the label column or list of columns.
            This is a list of columns for multilabel models.
    :type target_column: str or list[str]
    :param task_type: The type of task to be performed.
    :type task_type: str
    :return: The list of extracted features and the feature names.
    :rtype: list, list
    '''
    results = []
    base_feature_names = ["positive_words", "negative_words",
                          "negation_words", "negated_entities",
                          "named_persons", "sentence_length"]
    single_text_col_tasks = [ModelTask.TEXT_CLASSIFICATION,
                             ModelTask.MULTILABEL_TEXT_CLASSIFICATION]
    if task_type in single_text_col_tasks:
        feature_names = base_feature_names
    elif task_type == ModelTask.QUESTION_ANSWERING:
        feature_names = []
        prefixes = [QuestionAnsweringFields.CONTEXT + "_",
                    QuestionAnsweringFields.QUESTION + "_"]
        for prefix in prefixes:
            for feature_name in base_feature_names:
                feature_names.append(prefix + feature_name)

        feature_names.append("question_type")
    else:
        raise ValueError("Unknown task type: {}".format(task_type))
    if not isinstance(target_column, list):
        target_column = [target_column]
    text_features = text_dataset.drop(target_column, axis=1)

    if task_type in single_text_col_tasks:
        sentences = text_features.iloc[:, 0].tolist()
        for sentence in tqdm(sentences):
            extracted_features = []
            add_extracted_features_for_sentence(sentence, extracted_features)
            results.append(extracted_features)
    elif task_type == ModelTask.QUESTION_ANSWERING:
        for _, row in tqdm(text_features.iterrows()):
            extracted_features = []
            add_extracted_features_for_sentence(
                row[QuestionAnsweringFields.CONTEXT], extracted_features)
            add_extracted_features_for_sentence(
                row[QuestionAnsweringFields.QUESTIONS], extracted_features, sentence_type="QUESTION")
            results.append(extracted_features)
    else:
        raise ValueError("Unknown task type: {}".format(task_type))
    return results, feature_names


def add_extracted_features_for_sentence(sentence, extracted_features, sentence_type=None):
    global nlp
    if nlp is None:
        nlp = spacy.load("en_core_web_sm")
        ts = termset("en")
        nlp.add_pipe("negex", config={"neg_termset": ts.get_patterns()})
    doc = nlp(sentence)
    positive_negative_count = exts.positive_negative_word_count(doc)
    named_persons = exts.get_named_persons(doc)
    neg_words_and_entities = exts.detect_negation_words_and_entities(doc)
    sentence_length = len(sentence)
    features = [positive_negative_count["positive_word_count"],
                positive_negative_count["negative_word_count"],
                neg_words_and_entities["negation_words"],
                neg_words_and_entities["negated_entities"],
                len(named_persons),
                sentence_length]

    if sentence_type == 'QUESTION':
        question_type = get_question_type(sentence)
        features.append(question_type)

    # TODO: This extractor seems to be very slow:
    # mf_count = exts.get_male_female_words_count(doc)

    extracted_features.extend(features)


def get_question_type(qtext):

    if re.search(r'\b\A(can|could|will|would|have|has|do|does|' +
                 'did|is|are|was|may|might)', qtext, re.I):
        return "YES/NO"
    elif re.search(r'\b\A(what|which)\s+(\w+)', qtext, re.I):
        nextword = re.search(r'\b\A(what|which)\s+(\w+)', qtext, re.I).group(2)
        if nextword == "year" or nextword == "month" or nextword == "date" or nextword == "day":
            return "WHEN"
        else:
            return "WHAT"
    elif re.search(r'\bwho\s', qtext, re.I):
        return "WHO"
    elif re.search(r'\bwhy\s', qtext, re.I):
        return "WHY"
    elif re.search(r'\bwhere\s', qtext, re.I):
        return "WHERE"
    elif re.search(r'\bhow\s', qtext, re.I):
        nextword = re.search(r'\b(how)\s(\w+)', qtext, re.I).group(2)
        if nextword == "many" or nextword == "much" or nextword == "long":
            return "NUMBER"
        else:
            return "HOW"
    elif re.search(r'\bwhen\s', qtext, re.I):
        return "WHEN"
    elif "in what year" in qtext or "in which year" in qtext:
        return "WHEN"
    elif re.search(r'\bto\swhom\s', qtext, re.I):
        return "WHO"
    else:
        return "OTHER"
